#include "stdafx.h"
#include "Image.h"


/*
#define DIB_ROW_BYTES(width,bpp) (((width)*(bpp)+31)/32*4)
*/
#define DIB_ROW_BYTES(width,bpp) ((((width)*(bpp)+31)>>5)<<2)




static void CropImage(const BITMAPINFO *pbmiSrc,const void *pSrcData,
						int Left,int Top,int Width,int Height,void *pDstData)
{
	int x,y;
	int SrcRowBytes,DstRowBytes;
	const BYTE *p;
	BYTE *q;

	SrcRowBytes=DIB_ROW_BYTES(pbmiSrc->bmiHeader.biWidth,
												pbmiSrc->bmiHeader.biBitCount);
	DstRowBytes=DIB_ROW_BYTES(Width,24);
	q=(BYTE*)pDstData+(Height-1)*DstRowBytes;
	for (y=0;y<Height;y++) {
		p=(const BYTE*)pSrcData+
			((pbmiSrc->bmiHeader.biHeight>0)?
				(pbmiSrc->bmiHeader.biHeight-1-(Top+y)):(Top+y))*SrcRowBytes+
										Left*pbmiSrc->bmiHeader.biBitCount/8;
		if (pbmiSrc->bmiHeader.biBitCount==24) {
			CopyMemory(q,p,Width*3);
		} else {
			BYTE *r;

			r=q;
			for (x=0;x<Width;x++) {
				*r++=*p++;
				*r++=*p++;
				*r++=*p++;
				p++;
			}
		}
		q-=DstRowBytes;
	}
}


HGLOBAL ResizeImage(const BITMAPINFO *pbmiSrc,const void *pSrcData,
									const RECT *pSrcRect,int Width,int Height)
{
	HGLOBAL hGlobal;
	BITMAPINFOHEADER *pbmihDst;
	void *pDstData;
	int *pnSrcPos;
	int SrcLeft,SrcTop,SrcWidth,SrcHeight;
	int SrcPlanes,SrcXCenter,SrcYCenter,DstXCenter,DstYCenter;
	int x,y,x1,y1,dx1,dy1,dx2,dy2,YOffset;
	int b,g,r;
	int SrcRowBytes,DstPadBytes;
	const BYTE *p,*p1;
	BYTE *q;

	if (pbmiSrc->bmiHeader.biBitCount!=24 && pbmiSrc->bmiHeader.biBitCount!=32)
		return NULL;
	hGlobal=GlobalAlloc(GMEM_MOVEABLE | GMEM_SHARE,
					sizeof(BITMAPINFOHEADER)+DIB_ROW_BYTES(Width,24)*Height);
	if (hGlobal==NULL)
		return NULL;
	pbmihDst=(BITMAPINFOHEADER*)GlobalLock(hGlobal);
	pbmihDst->biSize=sizeof(BITMAPINFOHEADER);
	pbmihDst->biWidth=Width;
	pbmihDst->biHeight=Height;
	pbmihDst->biPlanes=1;
	pbmihDst->biBitCount=24;
	pbmihDst->biCompression=BI_RGB;
	pbmihDst->biSizeImage=0;
	pbmihDst->biXPelsPerMeter=0;
	pbmihDst->biYPelsPerMeter=0;
	pbmihDst->biClrUsed=0;
	pbmihDst->biClrImportant=0;
	pDstData=pbmihDst+1;
	if (pSrcRect!=NULL) {
		SrcLeft=pSrcRect->left;
		SrcTop=pSrcRect->top;
		SrcWidth=pSrcRect->right-pSrcRect->left;
		SrcHeight=pSrcRect->bottom-pSrcRect->top;
	} else {
		SrcLeft=SrcTop=0;
		SrcWidth=pbmiSrc->bmiHeader.biWidth;
		SrcHeight=abs(pbmiSrc->bmiHeader.biHeight);
	}
	if (SrcWidth==Width && SrcHeight==Height) {
		CropImage(pbmiSrc,pSrcData,SrcLeft,SrcTop,Width,Height,pDstData);
		GlobalUnlock(hGlobal);
		return hGlobal;
	}
	pnSrcPos=new int[Width];
	SrcPlanes=pbmiSrc->bmiHeader.biBitCount/8;
	SrcXCenter=((SrcWidth-1)<<8)/2;
	SrcYCenter=((SrcHeight-1)<<8)/2;
	DstXCenter=((Width-1)<<8)/2;
	DstYCenter=((Height-1)<<8)/2;
	for (x=0;x<Width;x++) {
		pnSrcPos[x]=((x<<8)-DstXCenter)*SrcWidth/Width+SrcXCenter;
		if (pnSrcPos[x]<0)
			pnSrcPos[x]=0;
		else if (pnSrcPos[x]>(SrcWidth-1)<<8)
			pnSrcPos[x]=(SrcWidth-1)<<8;
	}
	SrcRowBytes=DIB_ROW_BYTES(pbmiSrc->bmiHeader.biWidth,
												pbmiSrc->bmiHeader.biBitCount);
	DstPadBytes=DIB_ROW_BYTES(Width,24)-Width*3;
	q=(BYTE*)pDstData;
	for (y=0;y<Height;y++) {
		y1=((y<<8)-DstYCenter)*SrcHeight/Height+SrcYCenter;
		if (y1<0)
			y1=0;
		else if (y1>(SrcHeight-1)<<8)
			y1=(SrcHeight-1)<<8;
		dy2=y1&0xFF;
		dy1=0x100-dy2;
		YOffset=dy2>0?SrcRowBytes:0;
		for (x=0;x<Width;x++) {
			x1=pnSrcPos[x];
			dx2=x1&0xFF;
			dx1=0x100-dx2;
			p=(const BYTE*)pSrcData+((y1>>8)+SrcTop)*SrcRowBytes+
												((x1>>8)+SrcLeft)*SrcPlanes;
			p1=p+((dx2+0xFF)>>8)*SrcPlanes;
			b=(p[0]*dx1+p1[0]*dx2)*dy1;
			g=(p[1]*dx1+p1[1]*dx2)*dy1;
			r=(p[2]*dx1+p1[2]*dx2)*dy1;
			p+=YOffset;
			p1+=YOffset;
			b+=(p[0]*dx1+p1[0]*dx2)*dy2;
			g+=(p[1]*dx1+p1[1]*dx2)*dy2;
			r+=(p[2]*dx1+p1[2]*dx2)*dy2;
			*q++=b>>16;
			*q++=g>>16;
			*q++=r>>16;
		}
		q+=DstPadBytes;
	}
	delete [] pnSrcPos;
	GlobalUnlock(hGlobal);
	return hGlobal;
}
